# =============================================================================
# Minet Twitter Tweet Search CLI Action
# =============================================================================
#
# Logic of the `tw tweet-search` action.
#
import casanova
from twitwi import (
    normalize_tweets_payload_v2,
    format_tweet_as_csv_row
)
from twitwi.constants import TWEET_FIELDS, TWEET_EXPANSIONS, TWEET_PARAMS
from twitter import TwitterHTTPError

from minet.cli.utils import LoadingBar
from minet.twitter import TwitterAPIClient
from minet.cli.twitter.utils import validate_query_boundaries

ITEMS_PER_PAGE = 100


def twitter_tweet_search_action(cli_args):
    validate_query_boundaries(cli_args)

    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version='2'
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=TWEET_FIELDS,
        total=cli_args.total
    )

    loading_bar = LoadingBar(
        'Retrieving tweets',
        total=enricher.total,
        unit='tweet'
    )

    for row, query in enricher.cells(cli_args.query, with_rows=True):

        kwargs = {'query': query, 'max_results': ITEMS_PER_PAGE, 'sort_order': cli_args.sort_order, 'expansions': ','.join(TWEET_EXPANSIONS), 'params': TWEET_PARAMS}

        loading_bar.print('Searching for "%s"' % query)
        loading_bar.inc('queries')

        if cli_args.start_time:
            kwargs['start_time'] = cli_args.start_time
        if cli_args.end_time:
            kwargs['end_time'] = cli_args.end_time
        if cli_args.since_id:
            kwargs['since_id'] = cli_args.since_id
        if cli_args.until_id:
            kwargs['until_id'] = cli_args.until_id

        route = ['tweets', 'search', 'all'] if cli_args.academic else ['tweets', 'search', 'recent']

        while True:
            try:
                result = client.call(route, **kwargs)
            except TwitterHTTPError as e:
                loading_bar.inc('errors')

                if e.e.code == 404:
                    enricher.writerow(row)
                else:
                    raise e

                continue

            # Empty response
            if result['meta']['result_count'] == 0 and 'next_token' in result['meta']:
                kwargs['next_token'] = result['meta']['next_token']
                continue

            normalized_tweets = normalize_tweets_payload_v2(result, collection_source='api')

            for normalized_tweet in normalized_tweets:
                loading_bar.update()

                addendum = format_tweet_as_csv_row(normalized_tweet)
                enricher.writerow(row, addendum)

            if 'next_token' in result['meta']:
                kwargs['next_token'] = result['meta']['next_token']
            else:
                break



# =============================================================================
# Minet BuzzSumo Constants
# =============================================================================
#
# General constants used by the buzzsumo package.
#
ARTICLES_CSV_HEADERS = [
    'id',
    'published_date',
    'updated_at',
    'url',
    'og_url',
    'thumbnail',
    'domain_name',
    'subdomain',
    'num_linking_domains',
    'total_shares',
    'total_facebook_shares',
    'facebook_comments',
    'facebook_shares',
    'facebook_likes',
    'love_count',
    'wow_count',
    'haha_count',
    'angry_count',
    'sad_count',
    'twitter_shares',
    'twitter_user_id',
    'pinterest_shares',
    'total_reddit_engagements',
    'alexa_rank',
    'evergreen_score',
    'evergreen_score2',
    'youtube_trending_score',
    'youtube_views',
    'youtube_likes',
    'youtube_dislikes',
    'youtube_comments',
    'article_amplifiers',
    'title',
    'display_title',
    'author_name',
    'language',
    'num_words',
    'article_types',
    'video_length',
]



# =============================================================================
# Minet BuzzSumo Limit CLI Action
# =============================================================================
#
# Logic of the `bz limit` action.
#
from termcolor import colored

from minet.utils import prettyprint_integer
from minet.cli.utils import die, register_retryer_logger
from minet.buzzsumo import BuzzSumoAPIClient
from minet.buzzsumo.exceptions import (
    BuzzSumoInvalidTokenError
)


def buzzsumo_limit_action(cli_args):
    register_retryer_logger()

    client = BuzzSumoAPIClient(cli_args.token)

    try:
        limit = client.limit()
    except BuzzSumoInvalidTokenError:
        die('Your BuzzSumo token is invalid!')

    print(
        'With your token, you can still make',
        colored(prettyprint_integer(limit), 'green'),
        'calls to the BuzzSumo API until the end of the month.'
    )



# =============================================================================
# Minet BuzzSumo API Client
# =============================================================================
#
# BuzzSumo API client.
#
from json import JSONDecodeError
from urllib.parse import quote

from minet.utils import RateLimiterState, rate_limited_method
from minet.web import (
    create_request_retryer,
    retrying_method,
    request_json
)

from minet.buzzsumo.formatters import format_article
from minet.buzzsumo.exceptions import (
    BuzzSumoInvalidTokenError,
    BuzzSumoOutageError,
    BuzzSumoBadRequestError,
    BuzzSumoInvalidQueryError
)

URL_TEMPLATE = 'https://api.buzzsumo.com%s?api_key=%s'
MAXIMUM_PAGE_NB = 98


def construct_url(endpoint, token, begin_timestamp=None, end_timestamp=None,
                  num_results=100, q=None, page=None):

    url = URL_TEMPLATE % (endpoint, token)

    url += '&num_results=%i' % num_results

    if begin_timestamp:
        url += '&begin_date=%s' % begin_timestamp
    if end_timestamp:
        url += '&end_date=%s' % (end_timestamp - 1)
    if q is not None:
        url += '&q=%s' % quote(q)
    if page is not None:
        url += '&page=%i' % page

    return url


def optimize_period_timestamps_wrt_nb_pages(period_timestamps, nb_pages, maximum_page_nb):
    new_period_timestamps = period_timestamps

    if any(nb_page > maximum_page_nb for nb_page in nb_pages):
        for i in range(len(nb_pages)):
            if nb_pages[i] > maximum_page_nb:
                new_period_timestamps.append((period_timestamps[i] + period_timestamps[i + 1]) / 2)

        new_period_timestamps.sort()

    return period_timestamps


class BuzzSumoAPIClient(object):
    def __init__(self, token):
        self.token = token
        self.retryer = create_request_retryer(
            additional_exceptions=[BuzzSumoOutageError]
        )

        # 10 calls per ~12s
        self.rate_limiter_state = RateLimiterState(10, 12)

    @retrying_method()
    @rate_limited_method()
    def request(self, url):
        try:
            err, response, data = request_json(url)
        except JSONDecodeError:
            raise BuzzSumoBadRequestError

        if err:
            raise err

        if response.status == 401:
            raise BuzzSumoInvalidTokenError

        if response.status == 406:
            raise BuzzSumoInvalidQueryError(data.get('error'), url=url, data=data)

        if response.status == 500:
            raise BuzzSumoBadRequestError

        if response.status > 500:
            raise BuzzSumoOutageError

        return response, data

    def limit(self):
        url = construct_url(
            '/search/articles.json',
            token=self.token,
            q='fake news',
            num_results=1
        )

        response, _ = self.request(url)

        return response.headers['X-RateLimit-Month-Remaining']

    def domain_summary(self, domain, begin_timestamp, end_timestamp):
        url = construct_url(
            '/search/articles.json',
            token=self.token,
            q=domain,
            begin_timestamp=begin_timestamp,
            end_timestamp=end_timestamp
        )

        _, data = self.request(url)

        return {
            'total_results': int(data['total_results']),
            'total_pages': data['total_pages']
        }

    def __get_nb_pages_per_period_dates(self, domain, period_timestamps):
        nb_pages = []

        for i in range(len(period_timestamps) - 1):
            info = self.domain_summary(
                domain,
                begin_timestamp=period_timestamps[i],
                end_timestamp=period_timestamps[i + 1]
            )

            nb_pages.append(info['total_pages'])

        return nb_pages

    def domain_articles(self, domain, begin_timestamp, end_timestamp):

        # Here we optimize the periods used to request the API, because BuzzSumo
        # prevents us from getting more than 99 pages.
        maximum_page_nb = MAXIMUM_PAGE_NB
        period_timestamps = [begin_timestamp, end_timestamp]
        nb_pages = [1000]

        # This loop creates adapted time periods that all return less than 99 pages of results:
        while any(nb_page > maximum_page_nb for nb_page in nb_pages):

            # We ask how many pages are needed to get all the articles for the given periods:
            nb_pages = self.__get_nb_pages_per_period_dates(domain, period_timestamps)

            # If a given period gets more than 98 pages, this period is then cut down in half:
            period_timestamps = optimize_period_timestamps_wrt_nb_pages(period_timestamps, nb_pages, maximum_page_nb)

        # Now we get all the results for the optimized periods
        for i in range(len(period_timestamps) - 1):
            page = 0

            while True:
                url = construct_url(
                    '/search/articles.json',
                    token=self.token,
                    q=domain,
                    begin_timestamp=period_timestamps[i],
                    end_timestamp=period_timestamps[i + 1],
                    page=page
                )

                _, data = self.request(url)

                if not data['results']:
                    break

                for article in data['results']:
                    yield format_article(article)

                page += 1



# =============================================================================
# Minet Twitter User Tweets CLI Action
# =============================================================================
#
# Logic of the `tw user-tweets` action.
#
import casanova
from datetime import datetime
from twitwi import (
    normalize_tweet,
    normalize_tweets_payload_v2,
    format_tweet_as_csv_row
)
from twitwi.constants import TWEET_FIELDS, TWEET_EXPANSIONS, TWEET_PARAMS
from twitter import TwitterHTTPError

from minet.cli.utils import LoadingBar
from minet.twitter.constants import TWITTER_API_MAX_STATUSES_COUNT
from minet.twitter import TwitterAPIClient
from minet.cli.twitter.utils import is_not_user_id, is_probably_not_user_screen_name

ITEMS_PER_PAGE = 100


def twitter_user_tweets_action(cli_args):

    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version='2' if cli_args.v2 else '1.1'
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=TWEET_FIELDS,
        total=cli_args.total
    )

    loading_bar = LoadingBar(
        'Retrieving tweets',
        total=enricher.total,
        unit='user'
    )

    for row, user in enricher.cells(cli_args.column, with_rows=True):
        max_id = None

        loading_bar.update_stats(user=user)

        min_date_reached = False

        kwargs = {}

        while True:
            if cli_args.v2:
                if is_not_user_id(user):
                    loading_bar.die('The column given as argument doesn\'t contain user ids, you have probably given user screen names as argument instead. With --api-v2, you can only use user ids to retrieve followers.')

                kwargs['max_results'] = ITEMS_PER_PAGE
                kwargs['params'] = TWEET_PARAMS
                kwargs['expansions'] = ','.join(TWEET_EXPANSIONS)

                if cli_args.exclude_retweets:
                    kwargs['exclude'] = 'retweets'
                if cli_args.min_date:
                    utc_dt = datetime.fromtimestamp(cli_args.min_date).strftime('%Y-%m-%d')
                    kwargs['start_time'] = utc_dt + 'T00:00:00Z'

                try:
                    tweets = client.call(['users', user, 'tweets'], **kwargs)
                except TwitterHTTPError as e:
                    loading_bar.print(e)
                    loading_bar.inc('errors')

                    if e.e.code == 404:
                        loading_bar.print('Could not find user "%s"' % user)
                    else:
                        loading_bar.print('An error happened when attempting to retrieve tweets from "%s" (HTTP status %i)' % (user, e.e.code))

                    break

                try:
                    normalized_tweets = normalize_tweets_payload_v2(tweets, collection_source='api')
                except TypeError:
                    loading_bar.print('An error happened when attempting to retrieve tweets from "%s"' % user)
                    loading_bar.print('details: "%s"' % tweets['errors'][0]['detail'])
                    break

                for normalized_tweet in normalized_tweets:
                    addendum = format_tweet_as_csv_row(normalized_tweet)
                    enricher.writerow(row, addendum)

                if 'next_token' in tweets['meta']:
                    kwargs['pagination_token'] = tweets['meta']['next_token']
                else:
                    break

            else:
                if cli_args.ids:
                    if is_not_user_id(user):
                        loading_bar.die('The column given as argument doesn\'t contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command.')

                    kwargs = {'user_id': user}

                else:
                    if is_probably_not_user_screen_name(user):
                        loading_bar.die('The column given as argument probably doesn\'t contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command.')
                        # force flag to add

                    kwargs = {'screen_name': user}

                kwargs['include_rts'] = not cli_args.exclude_retweets
                kwargs['count'] = TWITTER_API_MAX_STATUSES_COUNT
                kwargs['tweet_mode'] = 'extended'

                if max_id is not None:
                    kwargs['max_id'] = max_id

                loading_bar.inc('calls')

                try:
                    tweets = client.call(['statuses', 'user_timeline'], **kwargs)
                except TwitterHTTPError as e:
                    loading_bar.inc('errors')

                    if e.e.code == 404:
                        loading_bar.print('Could not find user "%s"' % user)
                    else:
                        loading_bar.print('An error happened when attempting to retrieve tweets from "%s" (HTTP status %i)' % (user, e.e.code))

                    break

                if not tweets:
                    break

                max_id = min(int(tweet['id_str']) for tweet in tweets) - 1

                for tweet in tweets:
                    tweet = normalize_tweet(
                        tweet,
                        collection_source='api'
                    )
                    addendum = format_tweet_as_csv_row(tweet)

                    if cli_args.min_date:
                        if int(addendum[1]) < cli_args.min_date:
                            min_date_reached = True
                            break

                    enricher.writerow(row, addendum)

            loading_bar.inc('tweets', len(tweets))

            if min_date_reached:
                break

        loading_bar.update()



# =============================================================================
# Minet Twitter Tweet Count CLI Action
# =============================================================================
#
# Logic of the `tw tweet-count` action.
#
import casanova
from twitter import TwitterHTTPError

from minet.cli.utils import LoadingBar
from minet.cli.twitter.utils import validate_query_boundaries
from minet.twitter import TwitterAPIClient

ITEMS_PER_PAGE = 100

COUNT_FIELDS = ['tweet_count']
GRANULARIZED_COUNT_FIELDS = ['start_time', 'end_time', 'tweet_count']


def twitter_tweet_count_action(cli_args):
    validate_query_boundaries(cli_args)

    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version='2'
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=GRANULARIZED_COUNT_FIELDS if cli_args.granularity is not None else COUNT_FIELDS,
        total=cli_args.total
    )

    loading_bar = LoadingBar(
        'Counting tweets',
        total=enricher.total,
        unit='query',
        unit_plural='queries'
    )

    for row, query in enricher.cells(cli_args.query, with_rows=True):

        kwargs = {'query': query}

        loading_bar.print('Counting tweets for "%s"' % query)

        # Because we are greedy, we want stuff from the beginning of Twitter
        if cli_args.academic and not cli_args.start_time:
            kwargs['start_time'] = '2006-01-01T00:00:00Z'

        if cli_args.start_time:
            kwargs['start_time'] = cli_args.start_time
        if cli_args.end_time:
            kwargs['end_time'] = cli_args.end_time
        if cli_args.since_id:
            kwargs['since_id'] = cli_args.since_id
        if cli_args.until_id:
            kwargs['until_id'] = cli_args.until_id

        kwargs['granularity'] = cli_args.granularity or 'day'

        route = ['tweets', 'counts', 'all'] if cli_args.academic else ['tweets', 'counts', 'recent']

        total_count = 0

        while True:
            try:
                result = client.call(route, **kwargs)
                loading_bar.inc('calls')
            except TwitterHTTPError as e:
                loading_bar.inc('errors')

                if e.e.code == 404:
                    enricher.writerow(row)
                else:
                    raise e

                continue

            for count in result['data']:
                total_count += count['tweet_count']

                if cli_args.granularity is not None:
                    addendum = [count['start'], count['end'], count['tweet_count']]
                    enricher.writerow(row, addendum)

            if 'next_token' in result['meta']:
                kwargs['next_token'] = result['meta']['next_token']
            else:
                break

        loading_bar.update()

        if cli_args.granularity is None:
            enricher.writerow(row, [total_count])



# =============================================================================
# Minet Twitter Tweets CLI Action
# =============================================================================
#
# Logic of the `tw tweets` action.
#
import casanova
from twitwi import (
    normalize_tweet,
    normalize_tweets_payload_v2,
    format_tweet_as_csv_row
)
from twitwi.constants import TWEET_FIELDS, TWEET_EXPANSIONS, TWEET_PARAMS
from twitter import TwitterHTTPError
from ebbe import as_chunks

from minet.cli.utils import LoadingBar
from minet.twitter import TwitterAPIClient


def twitter_tweets_action(cli_args):

    client = TwitterAPIClient(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key,
        api_version='2' if cli_args.v2 else '1.1'
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=TWEET_FIELDS,
        total=cli_args.total
    )

    loading_bar = LoadingBar(
        'Retrieving tweets metadata',
        total=enricher.total,
        unit='tweet'
    )

    def call_client(ids):
        if cli_args.v2:
            kwargs = {'ids': ids, 'expansions': ','.join(TWEET_EXPANSIONS), 'params': TWEET_PARAMS}
            return client.call(['tweets'], **kwargs)
        else:
            kwargs = {'_id': ids, 'tweet_mode': 'extended'}
            return client.call(['statuses', 'lookup'], **kwargs)

    for chunk in as_chunks(100, enricher.cells(cli_args.column, with_rows=True)):
        tweets = ','.join(row[1] for row in chunk)

        try:
            result = call_client(tweets)
        except TwitterHTTPError as e:
            loading_bar.inc('errors')

            if e.e.code == 404:
                for row, tweet in chunk:
                    enricher.writerow(row)
            else:
                raise e

            continue

        indexed_result = {}

        if cli_args.v2:
            normalized_tweets = normalize_tweets_payload_v2(result, collection_source='api')
            for normalized_tweet in normalized_tweets:
                addendum = format_tweet_as_csv_row(normalized_tweet)
                indexed_result[normalized_tweet['id']] = addendum

        else:
            for tweet in result:
                tweet = normalize_tweet(tweet, collection_source='api')
                addendum = format_tweet_as_csv_row(tweet)
                indexed_result[tweet['id']] = addendum

        for row, tweet in chunk:
            addendum = indexed_result.get(tweet)

            enricher.writerow(row, addendum)

        loading_bar.update(len(chunk))



# =============================================================================
# Minet BuzzSumo Domain Summary CLI Action
# =============================================================================
#
# Logic of the `bz domain-summary` action.
#
import casanova

from minet.cli.utils import LoadingBar
from minet.buzzsumo import BuzzSumoAPIClient
from minet.buzzsumo.exceptions import (
    BuzzSumoInvalidQueryError,
    BuzzSumoInvalidTokenError
)

SUMMARY_HEADERS = [
    'total_results',
    'total_pages'
]


def buzzsumo_domain_summary_action(cli_args):

    client = BuzzSumoAPIClient(cli_args.token)

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        add=SUMMARY_HEADERS,
    )

    loading_bar = LoadingBar(
        desc='Retrieving domain summary',
        unit='domain',
        total=enricher.total
    )

    for row, domain_name in enricher.cells(cli_args.column, with_rows=True):

        try:
            data = client.domain_summary(domain_name, cli_args.begin_date, cli_args.end_date)
        except BuzzSumoInvalidTokenError:
            loading_bar.die('Your BuzzSumo token is invalid!')
        except BuzzSumoInvalidQueryError as e:
            loading_bar.die('Invalid query: %s' % e.url + '\nMessage from the API: %s' % e)

        enricher.writerow(row, [data['total_results'], data['total_pages']])
        loading_bar.update()



# =============================================================================
# Minet Twitter CLI Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the twitter actions.
#
import casanova
import re
from twitter import TwitterHTTPError

from minet.cli.utils import LoadingBar
from minet.cli.exceptions import InvalidArgumentsError
from minet.twitter import TwitterAPIClient

CHARACTERS = re.compile(r'[A-Za-z_]')
NUMBERS = re.compile(r'[0-9]+')
TWITTER_SCREEN_NAME = re.compile(r'[a-zA-Z0-9_]{1,15}')

ITEMS_PER_PAGE = 1000


def validate_query_boundaries(cli_args):
    if cli_args.start_time is not None and cli_args.end_time is not None:
        if cli_args.end_time < cli_args.start_time:
            raise InvalidArgumentsError('--end-time should be after --start-time!')

    if cli_args.since_id and cli_args.until_id:
        if cli_args.until_id < cli_args.since_id:
            raise InvalidArgumentsError('--until-id should be greater than --since-id!')


def make_twitter_action(method_name, csv_headers):

    def action(cli_args):
        enricher = casanova.batch_enricher(
            cli_args.file,
            cli_args.output,
            keep=cli_args.select,
            add=csv_headers
        )

        loading_bar = LoadingBar(
            desc='Retrieving ids',
            unit=method_name[:-1],
            stats={
                'users': 0
            }
        )

        client = TwitterAPIClient(
            cli_args.access_token,
            cli_args.access_token_secret,
            cli_args.api_key,
            cli_args.api_secret_key,
            api_version='2' if cli_args.v2 else '1.1'
        )

        resuming_state = None

        if cli_args.resume:
            resuming_state = cli_args.output.pop_state()

        for row, user in enricher.cells(cli_args.column, with_rows=True):
            loading_bar.update_stats(user=user)

            all_ids = []
            next_cursor = -1
            result = None

            if cli_args.v2:
                next_cursor = None

            if resuming_state is not None and resuming_state.last_cursor:
                next_cursor = int(resuming_state.last_cursor)

            if cli_args.v2:
                if is_not_user_id(user):
                    loading_bar.die('The column given as argument doesn\'t contain user ids, you have probably given user screen names as argument instead. With --api-v2, you can only use user ids to retrieve followers.')

                client_kwargs = {'max_results': ITEMS_PER_PAGE}

            elif cli_args.ids:
                if is_not_user_id(user):
                    loading_bar.die('The column given as argument doesn\'t contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command.')

                client_kwargs = {'user_id': user}

            else:
                if is_probably_not_user_screen_name(user):
                    loading_bar.die('The column given as argument probably doesn\'t contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command.')
                    # force flag to add

                client_kwargs = {'screen_name': user}

            while True:

                skip_in_output = None

                if resuming_state:
                    skip_in_output = resuming_state.values_to_skip
                    resuming_state = None

                if not cli_args.v2:
                    client_kwargs['cursor'] = next_cursor

                    try:
                        result = client.call([method_name, 'ids'], **client_kwargs)
                    except TwitterHTTPError as e:

                        # The user does not exist
                        loading_bar.inc('users_not_found')
                        break

                    if result is not None:
                        all_ids = result.get('ids', [])
                        next_cursor = result.get('next_cursor', 0)

                        loading_bar.update(len(all_ids))

                        batch = []

                        for user_id in all_ids:
                            if skip_in_output and user_id in skip_in_output:
                                continue

                            batch.append([user_id])

                    else:
                        break

                else:
                    if method_name == 'friends':
                        method_name_v2 = 'following'
                    else:
                        method_name_v2 = method_name

                    try:
                        result = client.call(route=['users', user, method_name_v2], **client_kwargs)
                    except TwitterHTTPError as e:

                        # The user does not exist
                        loading_bar.inc('users_not_found')
                        break

                    if result is not None and 'data' in result:
                        batch = []

                        for follower_metadata in result['data']:
                            user_id = follower_metadata['id']

                            if skip_in_output and user_id in skip_in_output:
                                continue
                            batch.append([user_id])

                        loading_bar.update(len(result['data']))

                        if 'next_token' in result['meta']:
                            next_cursor = result['meta']['next_token']
                            client_kwargs['pagination_token'] = next_cursor
                        else:
                            next_cursor = None

                    else:
                        break

                enricher.writebatch(row, batch, next_cursor or None)

                if next_cursor is None or next_cursor == 0:
                    break

            loading_bar.inc('users')

    return action


def is_not_user_id(item):
    return bool(re.match(CHARACTERS, item))


def is_probably_not_user_screen_name(item):
    matches = TWITTER_SCREEN_NAME.fullmatch(item)
    if matches:
        return bool(NUMBERS.fullmatch(item))
    return True
